LoadAddOn("LibDataBroker-1.1")

local _G = _G
local RW = _G.RW3

local expect = RW.Debug.expect
local DEBUG = RW.DEBUG

local IsShiftKeyDown, IsAltKeyDown, IsControlKeyDown = IsShiftKeyDown, IsAltKeyDown, IsControlKeyDown
local pairs = pairs

local LDB = {}
RW.Common:NamedTable(LDB, "LDB")
RW.LDB = LDB

local ldb = LibStub("LibDataBroker-1.1")

local LDBCache = {}
local LDBIndex = {}

local ValidClickTypes = {
	["Left"] = true,
	["LeftShift"]  = true,
	["LeftAlt"] = true,
	["LeftCtrl"] = true,
	["Right"] = true,
	["RightShift"]  = true,
	["RightAlt"] = true,
	["RightCtrl"] = true,
}


local function OnTooltipShow(self, tt)
	tt:AddLine("Raid Watch 3")
	
	for id, line in pairs(self.TooltipLines) do
		tt:AddLine(line)
	end	
	
	for i, line in pairs(self.ClickHelpLines) do
		tt:AddLine(line)
	end
end


local function CallClickFunctions(functions)
	if not functions then return end
	if DEBUG then
		expect(functions, "typeof", "table")
	end
	for ID, func in pairs(functions) do
		func()
	end
end

local function OnClick(self, frame, button)
	local ShiftKeyIsDown = IsShiftKeyDown()
	local AltKeyIsDown = IsAltKeyDown()
	local CtrlKeyIsDown = IsControlKeyDown()
	
	if button == "LeftButton" then
		if ShiftKeyIsDown then
			CallClickFunctions(self.ClickFunctions["LeftShift"])
		elseif AltKeyIsDown then
			CallClickFunctions(self.ClickFunctions["LeftAlt"])
		elseif AltKeyIsDown then
			CallClickFunctions(self.ClickFunctions["LeftCtrl"])
		else
			CallClickFunctions(self.ClickFunctions["Left"])
		end
		
	elseif button == "RightButton" then
		if ShiftKeyIsDown then
			CallClickFunctions(self.ClickFunctions["RightShift"])
		elseif AltKeyIsDown then
			CallClickFunctions(self.ClickFunctions["RightAlt"])
		elseif AltKeyIsDown then
			CallClickFunctions(self.ClickFunctions["RightCtrl"])
		else
			CallClickFunctions(self.ClickFunctions["Right"])
		end
	end
end


function LDB:Init()
	RW.UnregisterCallback(self, "Enabled")
end

function LDB:RegisterLDB(Name, LDBType)
	if DEBUG then
		expect(Name, "typeof", "string")
		expect(Name, "not_inset", LDBCache)
		expect(LDBType, "typeof", "string")
		expect(LDBType, "inset", "data source;launcher")
	end
	if not ldb then return end
	
	local obj = setmetatable({
		TooltipLines = {},
		ClickHelpLines = {},
		ClickFunctions = {},
	}, {__index = LDBIndex})	
	
	for t in pairs(ValidClickTypes) do
		obj.ClickFunctions[t] = {}
	end
	
	local l = ldb:NewDataObject(Name)
	l.OnTooltipShow = function(tt)
		OnTooltipShow(obj, tt)
	end
	l.OnClick = function(self, button)
		OnClick(obj, self, button)
	end
	obj.object = l
	
	tinsert(LDBCache, obj)
	
	return obj
end

--- Set the icon of the LDB object
-- @param icon Icon to set
-- @usage obj:SetIcon("Interface\\ICONS\\INV_Misc_EngGizmos_01")
function LDBIndex:SetIcon(icon)
	if DEBUG then
		expect(icon, "typeof", "string")
	end
	self.object.icon = icon
end

--- Set the sufix part of the LDB object
-- @param suffix The suffix
-- @usage obj:SetSuffix("str")
function LDBIndex:SetSuffix(suffix)
	if DEBUG then
		expect(suffix, "typeof", "string;number")
	end
	self.object.suffix = suffix	
end

--- Set the value part of the LDB object
-- @param value The value
-- @usage obj:SetValue(546)
function LDBIndex:SetValue(value)
	if DEBUG then
		expect(value, "typeof", "string;number")
	end
	self.object.value = value
end

--- Set the text part of the LDB object
-- @param text The text
-- @usage obj:SetText("My Addon")
function LDBIndex:SetText(text)
	if DEBUG then
		expect(text, "typeof", "string;number")
	end
	self.object.text = text
end

--- Set the label part of the LDB object
-- @param label The label
-- @usage obj:SetLabel("Active")
function LDBIndex:SetLabel(label)
	if DEBUG then
		expect(label, "typeof", "string")
	end
	self.object.label = label
end

--- Add a click function
-- @param clickType Button + modifier combination. Left, LeftShift, LeftAlt, LeftCtrl, Right, RightShift...
-- @param ID Identification of this function, used to remove it later.
-- @param func Function to call
-- @usage obj:AddClickFunction("LeftShift", "MyLeftShiftFunc", function() --[[ Do stuff when shift + left button is pressed ]] end)
function LDBIndex:AddClickFunction(clickType, ID, func)
	if DEBUG then
		expect(clickType, "typeof", "string")
		expect(ID, "typeof", "string;number")
		expect(func, "typeof", "function")
	end
	
	if not ValidClickTypes[clickType] then
		error(("Bad clickType to `AddClickFunction': %q"):format(clickType), 2)
	end
	if self.ClickFunctions[clickType][ID] then
		error(("Click function for %q with ID %q allready exists"):format(clickType, ID), 2)
	end

	self.ClickFunctions[clickType][ID] = func
end

--- Remove a click function
-- @param clickType Button + modifier combination. Left, LeftShift, LeftAlt, LeftCtrl, Right, RightShift...
-- @param ID Identification of the function to remove, same as supplied when adding it.
-- @usage obj:RemoveClickFunction("LeftShift", "MyLeftShiftFunc")
function LDBIndex:RemoveClickFunction(clickType, ID)
	if DEBUG then
		expect(clickType, "typeof", "string")
		expect(ID, "typeof", "string;number")
	end
	
	if not ValidClickTypes[clickType] then
		error(("Bad clickType to `RemoveClickFunction': %q"):format(clickType), 2)
	end

	self.ClickFunctions[clickType][ID] = nil
end

--- Add a special tooltip line to explain click events
-- @param ID Identification of the lin
-- @line Text to display on that line
-- @param clickType Button + modifier combination. Left, LeftShift, LeftAlt, LeftCtrl, Right, RightShift...
-- @usage obj:AddClickHelpLine("LeftClikcHelp", "Open config", "Left")
function LDBIndex:AddClickHelpLine(ID, line, clickType)
	if DEBUG then
		expect(line, "typeof", "string;number")
		expect(line, "typeof", "string;number")
	end
	
	self.ClickHelpLines[ID] = line
end

--- Remove a tootip click help line
-- @param ID Identification of the line to remove.
-- @usage obj:RemoveClickHelpLine("LeftClikcHelp")
function LDBIndex:RemoveClickHelpLine(ID)
	if DEBUG then
		expect(ID, "typeof", "string;number")
	end
	
	self.ClickHelpLines[ID] = nil
end

--- Add a line to the tooltip
-- @param ID Identification of the line
-- @param line Text show on the line.
-- @usage obj:AddTooltipLine("Line1", "Welcome")
function LDBIndex:AddTooltipLine(ID, line)
	if DEBUG then
		expect(ID, "typeof", "string")
		expect(line, "typeof", "string;number")
	end
	
	self.TooltipLines[ID] = line
end

--- Remove a line from the tooltip
-- @param ID Identification of the line to remove.
-- @usage obj:RemoveTooltipLine("Line1")
function LDBIndex:RemoveTooltipLine(ID)
	if DEBUG then
		expect(ID, "typeof", "string")
	end
	
	self.TooltipLines[ID] = nil
end

RW.RegisterCallback(LDB, "Enabled", "Init")