local name, addon_table = ...

-- Up Values
local _G = _G
local RW = _G.RW3
local DEBUG = RW.DEBUG
local Debug = RW.Debug
local expect = Debug.expect

local cache = {}
local active = {}
local ObjectCount = {}
Debug:AddGlobal("ControlCache", cache)
Debug:AddGlobal("ControlActive", active)

local Objects = {}
RW.Objects = Objects

local ObjectPrototype = {}
local SpecialObjectProtoypes = {}
local DeleteFuncs = {}
local RetrieveFuncs = {}

local frame = CreateFrame("Frame")
local AnimCache = frame:CreateAnimationGroup()
local DefaultFont, DefaultFontSize = GameFontNormal:GetFont()

local function CallObjectFunc(object, functionName, ...)
	if object[functionName] then
		object[functionName](object, ...)
	end
end

local function CreateObject(otype, name, parent, template, ...)
	if otype == "Texture" then
		return parent:CreateTexture(name, "BACKGROUND", 0)
	end
	if otype == "FontString" then
		return parent:CreateFontString(name, "BACKGROUND", template)
	end
	if otype == "Alpha" or otype == "Rotation" or otype == "Scale" or otype == "Translation" then
		return paremt:CreateAnimation(otype, name, template, ...)
	end
	return CreateFrame(otype, name, parent, template, ...)
end

local function GetCacheObject(otype, customType, parent, _OnCreate, template, ...)
	ocache = cache[otype]
	if not ocache then
		ocache = {}
		cache[otype] = ocache
		ObjectCount[otype] = 1
	end
	
	local control = next(ocache)
	if not control then
		local name = ("RaidWatch3_%s_%d"):format(otype, ObjectCount[otype])
		control = CreateObject(otype, name, parent, template, ...)
	else
		ocache[control] = nil
		active[otype][control] = control
		return control
	end
	
	for fname, func in pairs(ObjectPrototype) do
		control[fname] = func
	end
	if SpecialObjectProtoypes[otype] then
		for fname, func in pairs(SpecialObjectProtoypes[otype]) do
			control[fname] = func
		end
	end
	
	if _OnCreate then
		_OnCreate(control)
	end
	
	active[otype] = active[otype] or {}
	active[otype][control] = control
	
	control.otype = otype
	
	return control
end

local function GetObject(otype, customType, parent, ...)
	local control = GetCacheObject(otype, customType, parent, ...)
		
	CallObjectFunc(control, "SetParent", parent)
	CallObjectFunc(control, "Show")
	CallObjectFunc(control, "ClearAllPoints")
	
	return control
end

function Objects:GetFrame(parent)
	if DEBUG then
		expect(parent, "typeof", "frame")
	end
	
	return GetObject("Frame", false, parent)
end

function Objects:GetStatusBar(parent)
	if DEBUG then
		expect(parent, "typeof", "frame")
	end
	
	return GetObject("StatusBar", false, parent)
end

function Objects:GetTexture(parent, layer, subLevel)
	if DEBUG then
		expect(parent, "typeof", "frame")
		expect(layer, "typeof", "string;nil")		
		expect(subLevel, "typeof", "number;nil")
		if layer then
			expect(layer, "inset", "BACKGROUND;BORDER;ARTWORK;OVERLAY;HIGHLIGHT")
		end
	end
	
	local control = GetObject("Texture", false, parent)
	RetrieveFuncs.Texture(control, layer, subLevel)
	
	return control
end

function Objects:GetFontString(parent, layer, inherit, subLevel)
	if DEBUG then
		expect(parent, "typeof", "frame")		
		expect(layer, "typeof", "string;nil")
		expect(inherit, "typeof", "string;nil")
		expect(subLevel, "typeof", "number;nil")
		if layer then
			expect(layer, "inset", "BACKGROUND;BORDER;ARTWORK;OVERLAY;HIGHLIGHT")
		end
		if inherit then
			expect(inherit, "inset", _G)
		end
	end
	
	local control = GetObject("FontString", false, parent)
	RetrieveFuncs.FontString(control, layer, inherit, subLevel)
	return control
end

-- Retrieve functions
function RetrieveFuncs:Texture(layer, subLevel)	
	self:SetDrawLayer(layer or "BACKGROUND", subLevel or 0)
end

function RetrieveFuncs:FontString(layer, template, subLevel)	
	self:SetDrawLayer(layer or "BACKGROUND", subLevel)
	local font, fontsize, flags = template and _G[template]:GetFont() or DefaultFont, DefaultFontSize, nil
	self:SetFont(font, fontsize, flags)
	self:SetFontObject(_G[template] or _G["GameFontNormal"])
end

-- Delete functions
function DeleteFuncs:Texture()
	self:SetTexture(nil)
	self:SetVertexColor(1, 1, 1, 1)
	self:SetDrawLayer("BACKGROUND", 0)
end

function DeleteFuncs:FontString()
	self:SetText()
	self:SetJustifyH("CENTER")
	self:SetJustifyV("MIDDLE")
	self:SetNonSpaceWrap(true)
	self:SetTextColor(1, 1, 1, 1)
	self:SetFont(DefaultFont, DefaultFontSize, nil)
	self:SetFontObject(nil)
end

-- Prototype functions

--- Delete an object
-- @name Object:Delete
function ObjectPrototype:Delete()
	local otype = self.otype
	
	if self._OnDelete then
		self:_OnDelete()
		self._OnDelete = nil
	end
	
	if DeleteFuncs[otype] then
		DeleteFuncs[otype](self)
	end
	
	CallObjectFunc(self, "ClearAllPoints")
	CallObjectFunc(self, "Hide")
	CallObjectFunc(self, "SetPoint", "BOTTOM", UIParent, "TOP", 0, 10)
	CallObjectFunc(self, "SetParent", otype == "Animation" and AnimCache or UIParent)
	CallObjectFunc(self, "SetAlpha", 1)
	CallObjectFunc(self, "SetScale", 1)
	CallObjectFunc(self, "SetWidth", 0)
	CallObjectFunc(self, "SetHeight", 0)
	
	active[otype][self] = nil
	cache[otype][self] = self
end

-- FontString prototype
local FontStringPrototype = {}
SpecialObjectProtoypes.FontString = FontStringPrototype

--- Set the font size
-- @name FontString:SetFontSize
-- @param size Font size
function FontStringPrototype:SetFontSize(size)
	if DEBUG then
		expect(size, "typeof", "number")
	end
	local font, _, flags = self:GetFont()
	self:SetFont(font, size, flags)
end
